#! /bin/bash
# $Id: test.sh,v 1.31 2002/02/10 20:51:18 gerhard Exp $
# Copyright Gerhard Rieger 2001, 2002
# Published under the GNU General Public License V.2, see file COPYING

# perform lots of tests on socat

# this script uses functions; you need a shell that supports them

#set -vx

SOCAT=./socat
opts="-d"
debug="1"

#E=-e	# Linux
if   [ $(echo "x\c") = "x" ]; then E=""
elif [ $(echo -e "x\c") = "x" ]; then E="-e"
else
    echo "cannot suppress trailing newline on echo" >&2
    exit 1
fi

TD="/tmp/$$"; td="$TD"
rm -rf "$TD" || (echo "cannot rm $TD" >&2; exit 1)
mkdir "$TD"
trap "rm -r $TD" 0 3

echo "using temp directory $TD"

# test if addresses are sorted alphabetically:
echo $E "testing if address array is sorted...\c"
TF="$TD/socat-?"
IFS="$(echo $E ' \n\t')"
"$SOCAT" -? |sed '1,/address-head:/ d' |egrep 'groups=' |while IFS="$IFS:" read x y; do echo "$x"; done >"$TF"
"$SOCAT" -? |sed '1,/address-head:/ d' |egrep 'groups=' |while IFS="$IFS:" read x y; do echo "$x"; done |sort |diff "$TF" - >"$TF-diff"
if [ -s "$TF-diff" ]; then
    echo $E "\n*** address array is not sorted. Wrong entries:" >&2
    /bin/cat "$TD/socat-?-diff" >&2
else
    echo " ok"
fi
#/bin/rm "$TF"
#/bin/rm "$TF-diff"

# test if address options array ("optionnames") is sorted alphabetically:
echo $E "testing if address options are sorted...\c"
TF="$TD/socat-??"
"$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=' |awk '{print($1);}' >"$TF"
"$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=' |awk '{print($1);}' |sort |diff "$TF" - >"$TF-diff"
if [ -s "$TF-diff" ]; then
    echo $E "\n*** option array is not sorted. Wrong entries:" >&2
    /bin/cat "$TD/socat-??-diff" >&2
else
    echo " ok"
fi
/bin/rm "$TF"
/bin/rm "$TF-diff"

#==============================================================================
if false; then

# inquire which options are available
OPTS_ANY=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*ANY' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_BLK=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*BLK' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_CHILD=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*CHILD' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_CHR=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*CHR' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_DEVICE=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*DEVICE' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_EXEC=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*EXEC' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_FD=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*FD' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_FIFO=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*FIFO' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_FORK=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*FORK' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_LISTEN=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*LISTEN' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_NAMED=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*NAMED' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_OPEN=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*OPEN' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_RETRY=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*RETRY' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_RANGE=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*RANGE' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_FILE=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*REG' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_UNIX=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*UNIX' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_SOCKET=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*SOCKET' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_TERMIOS=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*TERMIOS' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_IP4=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*IP4' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_IP6=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*IP6' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_TCP=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*TCP' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_UDP=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*UDP' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_SOCKS4=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*SOCKS4' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_PROCESS=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*PROCESS' |awk '{print($1);}' |xargs echo |tr ' ' ',')
OPTS_APPL=$("$SOCAT" -?? |sed '1,/opt:/ d' |egrep 'groups=([A-Z]+,)*APPL' |awk '{print($1);}' |xargs echo |tr ' ' ',')


if [ $(id -u) = 0 ]; then
  _UID=nobody
  _GID=staff
else
  _UID=$(id -u)
  _GID=$(id -g)
fi

filloptionvalues() {
    local OPTS=",$1,"
    case "$OPTS" in
    *,user,*) OPTS=$(echo "$OPTS" |sed "s/,user,/,user=$_UID,/g");;
    esac
    case "$OPTS" in
    *,user-early,*) OPTS=$(echo "$OPTS" |sed "s/,user-early,/,user-early=$_UID,/g");;
    esac
    case "$OPTS" in
    *,user-late,*) OPTS=$(echo "$OPTS" |sed "s/,user-late,/,user-late=$_UID,/g");;
    esac
    case "$OPTS" in
    *,owner,*) OPTS=$(echo "$OPTS" |sed "s/,owner,/,owner=$_UID,/g");;
    esac
    case "$OPTS" in
    *,uid,*) OPTS=$(echo "$OPTS" |sed "s/,uid,/,uid=$_UID,/g");;
    esac
    case "$OPTS" in
    *,uid-l,*) OPTS=$(echo "$OPTS" |sed "s/,uid-l,/,uid-l=$_UID,/g");;
    esac
    case "$OPTS" in
    *,setuid,*) OPTS=$(echo "$OPTS" |sed "s/,setuid,/,setuid=$_UID,/g");;
    esac
    case "$OPTS" in
    *,group,*) OPTS=$(echo "$OPTS" |sed "s/,group,/,group=$_GID,/g");;
    esac
    case "$OPTS" in
    *,group-early,*) OPTS=$(echo "$OPTS" |sed "s/,group-early,/,group-early=$_GID,/g");;
    esac
    case "$OPTS" in
    *,group-late,*) OPTS=$(echo "$OPTS" |sed "s/,group-late,/,group-late=$_GID,/g");;
    esac
    case "$OPTS" in
    *,gid,*) OPTS=$(echo "$OPTS" |sed "s/,gid,/,gid=$_GID,/g");;
    esac
    case "$OPTS" in
    *,gid-l,*) OPTS=$(echo "$OPTS" |sed "s/,gid-l,/,gid-l=$_GID,/g");;
    esac
    case "$OPTS" in
    *,setgid,*) OPTS=$(echo "$OPTS" |sed "s/,setgid,/,setgid=$_GID,/g");;
    esac
    case "$OPTS" in
    *,mode,*) OPTS=$(echo "$OPTS" |sed "s/,mode,/,mode=0700,/g");;
    esac
    case "$OPTS" in
    *,perm,*) OPTS=$(echo "$OPTS" |sed "s/,perm,/,perm=0700,/g");;
    esac
    case "$OPTS" in
    *,perm-early,*) OPTS=$(echo "$OPTS" |sed "s/,perm-early,/,perm-early=0700,/g");;
    esac
    case "$OPTS" in
    *,perm-late,*) OPTS=$(echo "$OPTS" |sed "s/,perm-late,/,perm-late=0700,/g");;
    esac
    case "$OPTS" in
    *,path,*) OPTS=$(echo "$OPTS" |sed "s/,path,/,path=.,/g");;
    esac
    case "$OPTS" in
    *,linger,*) OPTS=$(echo "$OPTS" |sed "s/,linger,/,linger=2,/g");;
    esac
    case "$OPTS" in
    *,rcvtimeo,*) OPTS=$(echo "$OPTS" |sed "s/,rcvtimeo,/,rcvtimeo=1,/g");;
    esac
    case "$OPTS" in
    *,sndtimeo,*) OPTS=$(echo "$OPTS" |sed "s/,sndtimeo,/,sndtimeo=1,/g");;
    esac
    case "$OPTS" in
    *,bindtodevice,*) OPTS=$(echo "$OPTS" |sed "s/,bindtodevice,/,bindtodevice=lo0,/g");;
    esac
    echo $OPTS >&2
    expr "$OPTS" : ',\(.*\),'
}
# OPTS_FIFO: nothing yet

# OPTS_CHR: nothing yet

# OPTS_BLK: nothing yet

# OPTS_REG: nothing yet

OPTS_SOCKET=",$OPTS_SOCKET,"
OPTS_SOCKET=$(expr "$OPTS_SOCKET" : ',\(.*\),')

N=1
#------------------------------------------------------------------------------

#method=open
#METHOD=$(echo "$method" |tr a-z A-Z)
#TEST="$METHOD on file accepts all its options"
#TF=$TD/file$N
#DA="$(date)"
#OPTGROUPS=$(./socat -? |fgrep " $method:" |sed 's/.*=//')
#for g in $(echo $OPTGROUPS |tr ',' ' '); do
#    eval "OPTG=\$OPTS_$(echo $g |tr a-z A-Z)";
#    OPTS="$OPTS,$OPTG";
#done
##echo $OPTS
#
#for o in $(filloptionvalues $OPTS|tr ',' ' '); do
#    echo testing if $METHOD accepts option $o
#    touch $TF
#    "$SOCAT" -d -!!$method:$TF,$o /dev/null,ignoreof </dev/null
#    rm -f $TF
#done

#------------------------------------------------------------------------------

# unnamed pipe
if false; then
for addr in pipe; do
    ADDR=$(echo "$addr" |tr a-z A-Z)
    TEST="unnamed $ADDR accepts all its options"
    OPTGROUPS=$(./socat -? |egrep " $addr[^:]" |sed 's/.*=//')
    OPTS=
    for g in $(echo $OPTGROUPS |tr ',' ' '); do
	eval "OPTG=\$OPTS_$(echo $g |tr a-z A-Z)";
	OPTS="$OPTS,$OPTG";
    done
    #echo $OPTS

    for o in $(filloptionvalues $OPTS|tr ',' ' '); do
	echo testing if unnamed $ADDR accepts option $o
	"$SOCAT" -d $addr,$o /dev/null </dev/null
    done
done
fi

#------------------------------------------------------------------------------

# test addresses on files

N=1
if false; then
for addr in create; do
    ADDR=$(echo "$addr" |tr a-z A-Z)
    TEST="$ADDR on file accepts all its options"
    TF=$TD/file$N
    OPTGROUPS=$(./socat -? |fgrep " $addr:" |sed 's/.*=//')
    OPTS=
    for g in $(echo $OPTGROUPS |tr ',' ' '); do
	eval "OPTG=\$OPTS_$(echo $g |tr a-z A-Z)";
	OPTS="$OPTS,$OPTG";
    done
    #echo $OPTS

    for o in $(filloptionvalues $OPTS|tr ',' ' '); do
	echo testing if $ADDR accepts option $o
	rm -f $TF
	"$SOCAT" -d -!!$addr:$TF,$o /dev/null,ignoreof </dev/null
	rm -f $TF
    done
done
fi
#------------------------------------------------------------------------------

if false; then
for addr in exec system; do
    ADDR=$(echo "$addr" |tr a-z A-Z)
    TEST="$ADDR accepts all its options"
    OPTGROUPS=$(./socat -? |fgrep " $addr:" |sed 's/.*=//')
    OPTS=
    for g in $(echo $OPTGROUPS |tr ',' ' '); do
	eval "OPTG=\$OPTS_$(echo $g |tr a-z A-Z)";
	OPTS="$OPTS,$OPTG";
    done
    #echo $OPTS

    for o in $(filloptionvalues $OPTS|tr ',' ' '); do
	echo testing if $ADDR accepts option $o
	"$SOCAT" -d $addr:/bin/true,$o /dev/null,ignoreof </dev/null
    done
done
fi

#------------------------------------------------------------------------------

if false; then
for addr in fd; do
    ADDR=$(echo "$addr" |tr a-z A-Z)
    TEST="$ADDR accepts all its options"
    OPTGROUPS=$(./socat -? |fgrep " $addr:" |sed 's/.*=//')
    OPTS=
    for g in $(echo $OPTGROUPS |tr ',' ' '); do
	eval "OPTG=\$OPTS_$(echo $g |tr a-z A-Z)";
	OPTS="$OPTS,$OPTG";
    done
    #echo $OPTS

    for o in $(filloptionvalues $OPTS|tr ',' ' '); do
	echo testing if $ADDR accepts option $o
	"$SOCAT" -d $addr:0,$o!!stdout /dev/null,ignoreof </dev/null
    done
done
fi

#------------------------------------------------------------------------------

# test addresses on files

N=1
if false; then
for addr in open gopen; do
    ADDR=$(echo "$addr" |tr a-z A-Z)
    TEST="$ADDR on file accepts all its options"
    TF=$TD/file$N
    OPTGROUPS=$(./socat -? |fgrep " $addr:" |sed 's/.*=//')
    OPTS=
    for g in $(echo $OPTGROUPS |tr ',' ' '); do
	eval "OPTG=\$OPTS_$(echo $g |tr a-z A-Z)";
	OPTS="$OPTS,$OPTG";
    done
    #echo $OPTS

    for o in $(filloptionvalues $OPTS|tr ',' ' '); do
	echo testing if $ADDR accepts option $o
	touch $TF
	"$SOCAT" -d -!!$addr:$TF,$o /dev/null,ignoreof </dev/null
	rm -f $TF
    done
done
fi

#------------------------------------------------------------------------------

# test named pipe

N=1
if false; then
for addr in pipe; do
    ADDR=$(echo "$addr" |tr a-z A-Z)
    TEST="$ADDR on file accepts all its options"
    TF=$TD/pipe$N
    OPTGROUPS=$(./socat -? |fgrep " $addr:" |sed 's/.*=//')
    OPTS=
    for g in $(echo $OPTGROUPS |tr ',' ' '); do
	eval "OPTG=\$OPTS_$(echo $g |tr a-z A-Z)";
	OPTS="$OPTS,$OPTG";
    done
    #echo $OPTS

    for o in $(filloptionvalues $OPTS|tr ',' ' '); do
	echo testing if $ADDR accepts option $o
	rm -f $TF
	# blocks with rdonly, wronly
	case "$o" in rdonly|wronly) o="$o,nonblock" ;; esac
	"$SOCAT" -d $addr:$TF,$o /dev/null </dev/null
	rm -f $TF
    done
done
fi
#------------------------------------------------------------------------------

if false; then
for addr in stdio; do
    ADDR=$(echo "$addr" |tr a-z A-Z)
    TEST="$ADDR accepts all its options"
    OPTGROUPS=$(./socat -? |fgrep " $addr" |sed 's/.*=//')
    OPTS=
    for g in $(echo $OPTGROUPS |tr ',' ' '); do
	eval "OPTG=\$OPTS_$(echo $g |tr a-z A-Z)";
	OPTS="$OPTS,$OPTG";
    done
    #echo $OPTS

    for o in $(filloptionvalues $OPTS|tr ',' ' '); do
	echo testing if $ADDR accepts option $o
	"$SOCAT" -d $addr,$o /dev/null,ignoreof </dev/null
    done
done
fi

#------------------------------------------------------------------------------

if false; then
for addr in stdin; do
    ADDR=$(echo "$addr" |tr a-z A-Z)
    TEST="$ADDR accepts all its options"
    OPTGROUPS=$(./socat -? |fgrep " $addr	" |sed 's/.*=//')
    OPTS=
    for g in $(echo $OPTGROUPS |tr ',' ' '); do
	eval "OPTG=\$OPTS_$(echo $g |tr a-z A-Z)";
	OPTS="$OPTS,$OPTG";
    done
    #echo $OPTS

    for o in $(filloptionvalues $OPTS|tr ',' ' '); do
	echo testing if $ADDR accepts option $o
	"$SOCAT" -d -u $addr,$o /dev/null </dev/null
    done
done
fi

#------------------------------------------------------------------------------

if false; then
for addr in stdout stderr; do
    ADDR=$(echo "$addr" |tr a-z A-Z)
    TEST="$ADDR accepts all its options"
    OPTGROUPS=$(./socat -? |fgrep " $addr	" |sed 's/.*=//')
    OPTS=
    for g in $(echo $OPTGROUPS |tr ',' ' '); do
	eval "OPTG=\$OPTS_$(echo $g |tr a-z A-Z)";
	OPTS="$OPTS,$OPTG";
    done
    #echo $OPTS

    for o in $(filloptionvalues $OPTS|tr ',' ' '); do
	echo testing if $ADDR accepts option $o
	"$SOCAT" -d -u /dev/null $addr,$o
    done
done
fi

#------------------------------------------------------------------------------
# REQUIRES ROOT

if false; then
for addr in ip4; do
    ADDR=$(echo "$addr" |tr a-z A-Z)
    TEST="$ADDR accepts all its options"
    OPTGROUPS=$(./socat -? |fgrep " $addr:" |sed 's/.*=//')
    OPTS=
    for g in $(echo $OPTGROUPS |tr ',' ' '); do
	eval "OPTG=\$OPTS_$(echo $g |tr a-z A-Z)";
	OPTS="$OPTS,$OPTG";
    done
    #echo $OPTS

    for o in $(filloptionvalues $OPTS|tr ',' ' '); do
	echo testing if $ADDR accepts option $o
	"$SOCAT" -d $addr:127.0.0.1:200 /dev/null,ignoreof </dev/null
    done
done
fi

#==============================================================================

#TEST="stdio accepts all options of GROUP_ANY"
#CMD="$SOCAT -d -,$OPTS_ANY /dev/null"
#$CMD
#if [ $? = 0 ]; then
#    echo "... test $N ($TEST) succeeded"
##    echo "CMD=$CMD"
#else
#    echo "*** test $N ($TEST) FAILED"
#    echo "CMD=$CMD"
#fi
#
#N=$((N+1))
##------------------------------------------------------------------------------
#
#TEST="exec accepts all options of GROUP_ANY and GROUP_SOCKET"
#CMD="$SOCAT -d exec:/bin/true,$OPTS_ANY,$OPTS_SOCKET /dev/null"
#$CMD
#if [ $? = 0 ]; then
#    echo "... test $N ($TEST) succeeded"
##    echo "CMD=$CMD"
#else
#    echo "*** test $N ($TEST) FAILED"
#    echo "CMD=$CMD"
#fi

#------------------------------------------------------------------------------

fi	# false

#==============================================================================

N=1
#==============================================================================

function testecho {
    local num="$1"
    local title="$2"
    local arg1="$3";	[ -z "$arg1" ] && arg1="-"
    local arg2="$4";	[ -z "$arg2" ] && arg2="echo"
    local opts="$5"
    local tf="$td/test$N.stdout"
    local te="$td/test$N.stderr"
    local tdiff="$td/test$N.diff"
    local da="$(date)"
    #local cmd="$SOCAT $opts $arg1 $arg2"
    echo $E "testing $title (test $num)... \c"
    #echo "$da" |$cmd >"$tf" 2>"$te"
    echo "$da" |$SOCAT $opts "$arg1" "$arg2" >"$tf" 2>"$te"
    if [ "$?" != 0 ]; then
	echo "FAILED: $SOCAT:"
	echo "$SOCAT $opts $arg1 $arg2"
	cat "$te"
    elif echo "$da" |diff "$tf" - >"$tdiff" 2>&1; then
	echo "ok"
	if [ -n "$debug" ]; then cat $te; fi
    else
	echo "FAILED: diff:"
	echo "$SOCAT $opts $arg1 $arg2"
	cat "$tdiff"
    fi
}


TEST="unidirectional throughput from stdin to stdout"
testecho "$N" "$TEST" "stdin" "stdout" "$opts -u"
N=$((N+1))


TEST="stdio with simple echo via internal pipe"
testecho "$N" "$TEST" "stdio" "pipe"
N=$((N+1))


TEST="short form of stdio ('-') with simple echo via internal pipe"
testecho "$N" "$TEST" "-" "pipe"
N=$((N+1))


TEST="splitted form of stdio ('stdin!!stdout') with simple echo via internal pipe"
testecho "$N" "$TEST" "stdin!!stdout" "pipe"
N=$((N+1))


TEST="short splitted form of stdio ('-!!-') with simple echo via internal pipe"
testecho "$N" "$TEST" "-!!-" "pipe"
N=$((N+1))


TEST="file descriptors with simple echo via internal pipe"
testecho "$N" "$TEST" "0!!1" "pipe"
N=$((N+1))


TEST="simple echo via named pipe"
tp="$td/pipe$N"
testecho "$N" "$TEST" "" "pipe:$tp"
N=$((N+1))


TEST="simple echo via named pipe, specified twice"
tp="$td/pipe$N"
testecho "$N" "$TEST" "" "pipe:$tp,nonblock!!pipe:$tp"
N=$((N+1))


TEST="simple echo via file"
tf="$td/file$N"
testecho "$N" "$TEST" "" "$tf,ignoreeof!!$tf"
N=$((N+1))


TEST="simple echo via exec of cat with socketpair"
testecho "$N" "$TEST" "" "exec:cat" "$opts"
N=$((N+1))


TEST="simple echo via system() of cat with socketpair"
testecho "$N" "$TEST" "" "system:cat"
N=$((N+1))


TEST="simple echo via exec of cat with pipes"
testecho "$N" "$TEST" "" "exec:cat,pipes"
N=$((N+1))


TEST="simple echo via system() of cat with pipes"
testecho "$N" "$TEST" "" "system:cat,pipes"
N=$((N+1))


TEST="simple echo via exec of cat with pseudo terminal"
testecho "$N" "$TEST" "" "exec:cat,pty,echo=0,opost=0"
N=$((N+1))


TEST="simple echo via system() of cat with pseudo terminal"
testecho "$N" "$TEST" "" "system:cat,pty,echo=0,opost=0"
N=$((N+1))


TEST="simple echo via system() of cat with pipes, non stdio"
testecho "$N" "$TEST" "" "system:cat>&9 <&8,pipes,fdin=8,fdout=9"
N=$((N+1))

TEST="echo via dual system() of cat"
testecho "$N" "$TEST" "system:cat>&6,fdin=6!!system:cat<&7,fdout=7"
N=$((N+1))

TEST="simple echo via raw IP protocol"
if [ $(id -u) -ne 0 ]; then
    echo "testing $TEST (test $N)... must be root"
else
    testecho "$N" "$TEST" "" "ip:127.0.0.1:254"
fi
N=$((N+1))


TEST="echo via self connection of TCP IP4 socket (works due to bug in Linux?)"
p=2345
ts="127.0.0.1:$tsl"
testecho "$N" "$TEST" "" "tcp:127.100.0.1:$p,sp=$p,bind=127.100.0.1"
N=$((N+1))


TEST="echo via self connection of UDP IP4 socket"
p=2345
testecho "$N" "$TEST" "" "udp:127.100.0.1:$p,sp=$p,bind=127.100.0.1"
N=$((N+1))


if $($SOCAT -V |fgrep '#define WITH_IP6 1' >/dev/null); then
TEST="echo via self connection of UDP IP6 socket"
    tf="$td/file$N"
    p=2367
    testecho "$N" "$TEST" "" "udp6:0:0:0:0:0:0:0:1:$p,sp=$p,bind=0:0:0:0:0:0:0:1"
    N=$((N+1))
#else
#    echo "testing $TEST (test $N)... FAILED: IP6 not implemented"
fi


TEST="echo via two unidirectional UDP IP4 sockets"
tf="$td/file$N"
p1=2378
p2=2379
testecho "$N" "$TEST" "" "udp:127.0.0.1:$p2,sp=$p1!!udp:127.0.0.1:$p1,sp=$p2"
N=$((N+1))


function testdual {
    local
}


TEST="echo via connection to UNIX domain socket"
tf="$td/test$N.stdout"
te="$td/test$N.stderr"
ts="$td/socket$N"
da=$(date)
CMD1="$SOCAT unix-listen:$ts PIPE"
CMD2="$SOCAT -!!- unix:$ts"
echo $E "testing $TEST (test $N)... \c"
$CMD1 </dev/null >$tf 2>"$te" &
bg=$!	# backgroup process id
#sleep 1
echo "$da" |$CMD2 >"$tf" 2>"$tf"
if [ $? -eq 0 ] && echo "$da" |diff "$tf" -; then
   echo "ok"
   if [ -n "$debug" ]; then cat $te; fi
else
   echo "FAILED"
fi
kill $bg 2>/dev/null
N=$((N+1))


TEST="echo via connection to TCP V4 socket"
tf="$td/test$N.stdout"
te="$td/test$N.stderr"
tsl=65534
ts="127.0.0.1:$tsl"
da=$(date)
CMD1="$SOCAT TCP-listen:$tsl PIPE"
CMD2="$SOCAT stdin!!stdout TCP:$ts"
echo $E "testing $TEST (test $N)... \c"
$CMD1 >"$tf" 2>"$te" &
sleep 1
echo "$da" |$CMD2 >>"$tf" 2>>"$te"
if [ $? -eq 0 ] && echo "$da" |diff "$tf" -; then
   echo "ok"
   if [ -n "$debug" ]; then cat $te; fi
else
   echo "FAILED"
fi
N=$((N+1))


TEST="echo via connection to TCP V6 socket"
tf="$td/test$N.stdout"
te="$td/test$N.stderr"
tsl=65533
ts="::1:$tsl"
da=$(date)
CMD1="$SOCAT TCP6-listen:$tsl PIPE"
CMD2="$SOCAT stdin!!stdout TCP6:$ts"
echo $E "testing $TEST (test $N)... \c"
$CMD1 >"$tf" 2>"$te" &
sleep 1
echo "$da" |$CMD2 >>"$tf" 2>>"$te"
if [ $? -eq 0 ] && echo "$da" |diff "$tf" -; then
   echo "ok"
   if [ -n "$debug" ]; then cat $te; fi
else
   echo "FAILED"
fi
N=$((N+1))


TEST="file opening with gopen"
tf1="$td/test$N.1.stdout"
tf2="$td/test$N.2.stdout"
te="$td/test$N.stderr"
tdiff="$td/test$N.diff"
da=$(date)
echo "$da" >$tf1
CMD="$SOCAT $tf1!!/dev/null /dev/null,ignoreeof!!-"
echo $E "testing $TEST (test $N)... \c"
$CMD >"$tf2" 2>"$te"
if [ $? -ne 0 ]; then
    echo "FAILED: $SOCAT:"
    echo "$CMD"
    cat "$te"
elif ! diff "$tf1" "$tf2" >"$tdiff"; then
    echo "FAILED: diff:"
    cat "$tdiff"
else
   echo "ok"
fi
N=$((N+1))


TEST="pipe opening with gopen for reading"
tp="$td/pipe$N"
tf="$td/test$N.stdout"
te="$td/test$N.stderr"
tdiff="$td/test$N.diff"
da=$(date)
CMD="$SOCAT $tp!!/dev/null /dev/null,ignoreeof!!$tf"
echo $E "testing $TEST (test $N)... \c"
#mknod $tp p	# no mknod p on FreeBSD
mkfifo $tp
$CMD >$tf 2>"$te" &
bg=$!	# backgroup process id
echo "$da" >"$tp"
if ! diff "$tf1" "$tf2" >"$tdiff"; then
    if [ -s "$te" ]; then
	echo "FAILED: $SOCAT:"
	echo "$CMD"
	cat "$te"
    else
	echo "FAILED: diff:"
	cat "$tdiff"
    fi
else
   echo "ok"
   if [ -n "$debug" ]; then cat $te; fi
fi
N=$((N+1))


#==============================================================================
#TEST="echo via 'connection' to UDP V4 socket"
#tf="$td/file$N"
#tsl=65534
#ts="127.0.0.1:$tsl"
#da=$(date)
#"$SOCAT" UDP-listen:$tsl PIPE &
#sleep 2
#echo "$da" |"$SOCAT" stdin!!stdout UDP:$ts >"$tf"
#if [ $? -eq 0 ] && echo "$da" |diff "$tf" -; then
#   echo "... test $N succeeded"
#else
#   echo "*** test $N FAILED"
#fi
#N=$((N+1))
#==============================================================================
# TEST 4 - simple echo via new file
#N=4
#tf="$td/file$N"
#tp="$td/pipe$N"
#da=$(date)
#rm -f "$tf.tmp"
#echo "$da" |"$SOCAT" - FILE:$tf.tmp,ignoreeof >"$tf"
#if [ $? -eq 0 ] && echo "$da" |diff "$tf" -; then
#   echo "... test $N succeeded"
#else
#   echo "*** test $N FAILED"
#fi


# end

#echo "removing $td"
#/bin/rm -ri "$td"

exit


TEST="transferring from one file to another with echo"
tf1="$td/file$N.input"
tf2="$td/file$N.output"
testecho "$N" "$TEST" "" "echo"
